/**********************************************************************/
/***                                                                  */
/***        Author: Luis Eduardo Gamboa Guzmn                        */
/***        Last Updated: June 15th 2005                              */
/***                                                                  */
/***        Facultad de Ingeniera Elctrica                          */
/***        Universidad Michoacana de San Nicolas de Hidalgo          */
/***                                                                  */
/**********************************************************************/

#include <stdlib.h>
#include <math.h>
#include <memory.h>
#include "encode.h"
#include "dsp.h"
#include "wav.h"
#include "psychoacoustics.h"

int rangeofquant[17]={0,1,1,3,7,15,31,63,127,255,511,
						1023,2047,4095,8191,16383,32767};

/*
	K: Number of spectral samples
	Kp: Number of non-zero spectral samples being passed.
	P: Bit pool available for mantissas (i.e. without scale factors)
	b: band

	return: 
		number of bits to allocate to band b
	
*/
int allocatebitstoband(double *spl,double *mask,int Kp,int P,int b)
{
	int Nb; /* Number of samples in sub-band b */
	double SMRb; /* Signal to Mask Ratio for sub-band b */
	double aux1,aux2;
	double f;

	Nb=getbandsize(b);

	SMRb=signaltomaskratio(spl,mask,b);

	//printf("%e ",SMRb);

	aux1=(double)P/(double)Kp;
	aux2=SMRb-(double)(Nb)*SMRb/(double)Kp;

	aux2=log(10.0)*aux2/(20.0*log(2.0));

	/*if(SMRb<0.0)
		return 0;
	else
	{
		aux1=SMRb/6.0;
		if(aux1-floor(aux1) > 0.5)
			return ceil(aux1)+1;
		return (int)SMRb/6.0 +1 ;
	}*/

	if(aux1+aux2<0.0)
		return 0;


	f=floor(aux1+aux2);
	if(Nb==1)
	{
		if(f<1.0)
			return 0;
		else
			return 1;
	}
	if(aux1+aux2-f>0.7)
		return (int)ceil(aux1+aux2);
	return (int)f;
}

/* Return:
	Length (in bytes) of result */
UWORD quantizesignal(double *mdct,double *fftr,double *ffti,double *mask,int Nfreq,BYTE *result)
{
	int currentbit,lower,upper;
	int i,j;
	double *spl;
	double *mdctscaled;
	double PE=0.0,ln2=log(2.0);
	double max,aux;
	struct hdrpcklegg pckhdr;
	struct hdrpckleggsize pckhdrsize;
	BYTE bitallocations[25];
	WORD scalefactorsind[25];
	int bitsperblock;

	/* Nfreq is the length of the mdct */
	switch(Nfreq)
	{
		case 128:
			bitsperblock=100;
			break;
		case 256:
			bitsperblock=200;
			break;
		case 512:
			bitsperblock=400;
			break;
		case 1024:
			bitsperblock=900;
			break;
		default:
			bitsperblock=150;
			break;
	}

	/* Sounds better to compute bit allocations by using the SPL 
	of the mdct against the mask :S */
	spl=mdctspl(mdct,Nfreq);
	//spl=fftspl(fftr,ffti,Nfreq);
	mdctscaled=(double *)malloc(sizeof(double)*Nfreq);

	/* Copy the coefficients */
	for(i=0;i<Nfreq;i++)
		mdctscaled[i]=mdct[i];

	/* Compute bit allocation and scale factors for each subband */
	for(i=0;i<25;i++)
	{
		bitallocations[i]=allocatebitstoband(spl,mask,Nfreq,bitsperblock,i);
		/* 15-bit (including sign) is the maximum allowed quantizer */
		if(bitallocations[i]>15)
			bitallocations[i]=15;
		/* Forbit 1 bit quantizer */
		if(bitallocations[i]==1)
			bitallocations[i]++;

		getbandlimits(i,&lower,&upper);
		for(max=0.0,j=lower;j<=upper;j++)
		{
			aux=fabs(mdctscaled[j]);
			if(aux>max)
				max=aux;
		}
		/* Select appropiate scalefactor according to the bits allocated */
		aux=max/(double)QUANTIZERRANGE(bitallocations[i]);
		for(j=SFUPPER;j>=SFLOWER;j--)
		{
			if(aux>=COMPUTESF(j))
				break;
		}
		if(j<SFLOWER)
			j=SFLOWER;
		//if(aux-COMPUTESF(j) > COMPUTESF(j+1)-aux)
		//	j++;
		if(j>SFUPPER)
			j=SFUPPER;

		/*if(aux<COMPUTESF(j) && bitallocations[i]!=0)
		{
			printf("Value: %e Sf: %e Bits: %d  Range: %d\n",aux,COMPUTESF(j),bitallocations[i],QUANTIZERRANGE(bitallocations[i]));
			dequantize(quantize(aux,j+1,bitallocations[i]),j+1,bitallocations[i]);
		}*/
		//printf("%e ",COMPUTESF(j));
		//printf("%d %e %e\n",bitallocations[i],COMPUTESF(j),max);
		scalefactorsind[i]=j;
		//printf("%d %d %d %d\n",i,getbandsize(i),pckhdr.bitallocations[i],pckhdr.bitallocations[i]*getbandsize(i));
	}
	//exit(1);

	/* currentbit is set so that header can go first, however the header is not fully
	computed yet */
	currentbit=(sizeof(struct hdrpckleggsize)+sizeof(struct hdrpcklegg))*8;
	/* Store scale factors, if there are no bits allocated to a given subband the scale
	factor is not stored */
	//bitallocations[24]=0;
	for(j=0;j<25;j++)
	{
		if(bitallocations[j]!=0)
		{
			currentbit=writebits(result,(DWORD)(scalefactorsind[j])&((0x1<<SCALEFACTORBITS)-1),SCALEFACTORBITS,currentbit);
			//printf("%x %d %e\n",(DWORD)(scalefactorsind[j])&((0x1<<SCALEFACTORBITS)-1),(DWORD)(scalefactorsind[j]),COMPUTESF((scalefactorsind[j])));
			//printf("%d %d\n",j,bitallocations[j]);
		}
	}
	//exit(1);
	/* Quantize each sub-band and store in result */
	for(i=0;i<25;i++)
	{
		getbandlimits(i,&lower,&upper);
		currentbit=quantizeband(result,mdctscaled,lower,upper,bitallocations[i],scalefactorsind[i],currentbit);
	}

	/* Computes perceptual entropy */
	/*for(i=0;i<Nfreq;i++)
	{
		PE+=log(1.0+sqrt(spl[i]/mask[i]))/ln2;
	}*/

	//printf("Total bits allocated: %d\n\n",currentbit);
	//printf("Perceptual entropy: %e %e\n\n",PE,PE/(double)Nfreq);

	free(spl);
	free(mdctscaled);
	
	if(currentbit%8)
		pckhdrsize.lengthofpacket=currentbit/8+1-sizeof(struct hdrpckleggsize); /* Wastes a few bits, damn! */
	else
		pckhdrsize.lengthofpacket=currentbit/8-sizeof(struct hdrpckleggsize);
	
	/* Store length of packet and bit allocations at the beginning of result */
	pckhdr.alloc0=bitallocations[0];
	pckhdr.alloc1=bitallocations[1];
	pckhdr.alloc2=bitallocations[2];
	pckhdr.alloc3=bitallocations[3];
	pckhdr.alloc4=bitallocations[4];
	pckhdr.alloc5=bitallocations[5];
	pckhdr.alloc6=bitallocations[6];
	pckhdr.alloc7=bitallocations[7];
	pckhdr.alloc8=bitallocations[8];
	pckhdr.alloc9=bitallocations[9];
	pckhdr.alloc10=bitallocations[10];
	pckhdr.alloc11=bitallocations[11];
	pckhdr.alloc12=bitallocations[12];
	pckhdr.alloc13=bitallocations[13];
	pckhdr.alloc14=bitallocations[14];
	pckhdr.alloc15=bitallocations[15];
	pckhdr.alloc16=bitallocations[16];
	pckhdr.alloc17=bitallocations[17];
	pckhdr.alloc18=bitallocations[18];
	pckhdr.alloc19=bitallocations[19];
	pckhdr.alloc20=bitallocations[20];
	pckhdr.alloc21=bitallocations[21];
	pckhdr.alloc22=bitallocations[22];
	pckhdr.alloc23=bitallocations[23];
	pckhdr.alloc24=bitallocations[24];
	pckhdr.extra1=0;
	pckhdr.extra2=0;
	memcpy(result,&pckhdrsize,sizeof(struct hdrpckleggsize));
	memcpy(&result[sizeof(struct hdrpckleggsize)],&pckhdr,sizeof(struct hdrpcklegg));

	return pckhdrsize.lengthofpacket;
}

int quantizeband(BYTE *result,double *mdct,int lower,int upper,int bits,WORD scalefactor,int startbit)
{
	int currentbit=startbit;
	int i;
	int bitstogo;
	DWORD data;


	for(i=lower;i<=upper;i++)
	{
		/* Store all coefficients */
		data=quantize(mdct[i],scalefactor,bits);
		bitstogo=bits;
		currentbit=writebits(result,data,bitstogo,currentbit);
	}
	return currentbit;
}

int writebits(BYTE *buffer,DWORD data,int numberofbits,int startbit)
{
	int currentbit=startbit;
	int currentbyte=currentbit/8;
	int currentbitinbyte=currentbit%8;
	int aux;
	int bitstogo=numberofbits;
	int bitindata=0;
	DWORD mask;
	/* Writes the bits to the stream */
	while(bitstogo>0)
	{
		/* Bits left in current byte */
		aux=8-currentbitinbyte;
		/* Mask with size of bits left */
		mask=(0x1<<aux)-1;
		buffer[currentbyte]=buffer[currentbyte] | (BYTE)(((data>>bitindata)&mask )<<currentbitinbyte);
		
		/* Data did not fit in the remaining of current byte */
		if(bitstogo-aux>=0)
		{
			currentbit+=aux;
			currentbyte++;
			currentbitinbyte=0;
			bitstogo-=aux;
			bitindata+=aux;
		}
		else /* Remaining data fit in byte */
		{
			currentbitinbyte+=bitstogo;
			currentbit+=bitstogo;
			bitstogo=0;
		}
	}
	return currentbit;
}

DWORD quantize(double number,WORD scalefactor,int Nbits)
{
	double sf,scaled;
	DWORD quantized,mask;
	double range=QUANTIZERRANGE(Nbits);
	double c,f;
	//int sign; /* For AAC like */

/*	if(Nbits==1)
	{
		if(number>0.0)
			return 0x0;
		return 0x1;
	}*/
	
	mask = (0x1<<Nbits)-1;
	sf = COMPUTESF(scalefactor);

	/** AAC like **/
	/*if(number<0.0)
		sign=1;
	else
		sign=0;
	scaled = pow(fabs(number)/sf,0.75)-0.0946;
	c = ceil(scaled);
	f = floor(scaled);
	switch(sign)
	{
		case 0: /* positive */
	/*		if(scaled>range)
				scaled=range;
			/* Proper round */
	/*		if(scaled-f > 0.5)
			{
				if(c<=range)
					scaled=c;
			}
			quantized=(DWORD)scaled;
		case 1: /* negative */
	/*		if(scaled>range+1)
				scaled=range+1;
			/* Proper round */
	/*		if(scaled-f > 0.5)
			{
				if(c<=range+1)
					scaled=c;
			}
			quantized=(DWORD)(scaled*-1.0);
	}

	/** end AAC like **/

	/* Scaling is a real within range to be quantized by an Nbit quantizer */
	scaled = number / sf;
	c = ceil(scaled);
	f = floor(scaled);
	if(scaled>0.5)
	{
		if(scaled>range)
			scaled=range;
		/* Proper round */
		if(scaled-f > 0.5)
		{
			if(c<=range)
				scaled=c;
		}
	}
	else if(scaled<-0.5)
	{
		if(scaled<-range-1)
			scaled=-range-1;
		if(scaled-c<-0.5)
		{
			if(-f<=range+1)
				scaled=f;
		}
	}
	quantized=(DWORD)scaled;
	//printf("%x(%d) %e %e %e %d\n",quantized,quantized,number,scaled,sf,scalefactor);

	//printf("%e\n",(int)(quantized)*sf);

	quantized=quantized & mask;

	return quantized;
}

double dequantize(DWORD quantized,WORD scalefactor,int Nbits)
{
	DWORD mask;
	double number;
	double sf;

/*	if(Nbits==1)
	{
		if(quantized==0x0)
			return 1.0;
		return -1.0;
	}*/

	mask = 0x1<<(Nbits-1);
	//printf("%x ",quantized);
	if(mask & quantized)
	{
		/* Negative number, fill with ones to the left */
		mask = (0x1<<(sizeof(DWORD)*8))-1;
		mask = mask << Nbits;
		quantized = quantized | mask;
	}
	sf = COMPUTESF(scalefactor);
	/** AAC like **/
	/*if((int)(quantized) >= 0)
		number = pow((double)(int)(quantized)+0.0946,1.33333333333);
	else
		number = pow((double)(int)(quantized)-0.0946,1.33333333333);

	/** end AAC like **/

	number = (double)(int)(quantized) * sf;

	//printf("%x %e %e\n",quantized,number,sf);
	return number;
}

double midrisevalue(DWORD data,int Nbits)
{
	double step;
	DWORD sign,Nbits1;
	DWORD bmask;
	double number;

	Nbits1=Nbits-1;
	sign=(0x1<<Nbits1)&data;
	step=2.0/(double)((0x1<<Nbits)-1);
	bmask=(0x1<<Nbits1)-1;
	number=1.0-step*(bmask-(data&bmask));

	if(sign)
		return -number;

	return number;
}

DWORD midrisequantize(double number,int Nbits)
{
	double step;
	DWORD sign,Nbits1;
	DWORD quantized=0x0;
	DWORD aux;
	
	Nbits1=Nbits-1;
	if(number<0.0)
	{
		sign=(0x1<<Nbits1);
		number=-1.0*number;
	}
	else
		sign=0x0;

	step=1.0/(double)((0x1<<Nbits)-1);

	aux=(DWORD)(number/step);
	if(aux%2==0)
		aux++;

	aux=aux/2;
	quantized=aux+sign;

	return quantized;
}

double midtreadvalue(DWORD data,int Nbits)
{
	double step;
	DWORD sign,Nbits1;
	DWORD bmask;
	double number;

	Nbits1=Nbits-1;
	sign=(0x1<<Nbits1)&data;
	step=1.0/(double)((0x1<<Nbits1)-1);
	bmask=(0x1<<Nbits1)-1;
	number=step*(data&bmask);

	if(sign)
		return -number;

	return number;
}

/* Will not work with 1 bit quantizer */
DWORD midtreadquantize(double number,int Nbits)
{
	double step;
	DWORD sign,Nbits1;
	DWORD quantized=0x0;
	DWORD aux;
	
	Nbits1=Nbits-1;
	if(number<0.0)
	{
		sign=(0x1<<Nbits1);
		number=-1.0*number;
	}
	else
		sign=0x0;

	step=0.5/(double)((0x1<<Nbits1)-1);

	aux=(DWORD)(number/step);
	if(aux%2==1)
		aux++;
	aux=aux/2;
	quantized=aux+sign;

	return quantized;
}


void dequantizesignal(BYTE *bitallocations,BYTE *buffer,int lengthofpacket,double *mdct)
{
	WORD scalefactorind[25];
	DWORD dataaux;
	int i;
	int lower,upper;
	int currentbit;
	
	/* Extracts scale factors from buffer */
	currentbit=0;
	for(i=0;i<25;i++)
	{
		/* Is band i encoded? */
		if(bitallocations[i]!=0)
		{
			/* Copies the scale factor */
			currentbit=readbits(buffer,&dataaux,SCALEFACTORBITS,currentbit);
			//printf("%x ",dataaux);
			if(dataaux & 0x1<<(SCALEFACTORBITS-1))
				dataaux=dataaux | (((0x1<<28)-1)<<SCALEFACTORBITS);
			scalefactorind[i]=dataaux;
			//printf("%x %d %x\n",scalefactorind[i],scalefactorind[i],dataaux);
			//printf("%d %d\n",i,bitallocations[i]);
		}
		else
		{
			/* Band not encoded, set scale factor to 0.0 */
			scalefactorind[i]=0x0;
		}
	}

	/* Dequantize each sub-band and store in mdct */
	for(i=0;i<25;i++)
	{
		getbandlimits(i,&lower,&upper);
		currentbit=dequantizeband(buffer,mdct,lower,upper,bitallocations[i],scalefactorind[i],currentbit);
	}

}

int dequantizeband(BYTE *buffer,double *normalizedbands,int lower,int upper,int bits,WORD scalefactor,int startbit)
{
	int currentbit=startbit;
	int currentbyte=currentbit/8;
	int currentbitinbyte=currentbit%8;
	int i;
	DWORD data;


	if(bits==0)
	{
		for(i=lower;i<=upper;i++)
		{
			normalizedbands[i]=0.0;
		}
		return currentbit;
	}
	else 
	{	/* Decode all coefficients */
		for(i=lower;i<=upper;i++)
		{
			currentbit=readbits(buffer,&data,bits,currentbit);
			normalizedbands[i]=dequantize(data,scalefactor,bits);
		}
	}
	return currentbit;
}

int readbits(BYTE *buffer,DWORD *data,int numberofbits,int startbit)
{
	int currentbit=startbit;
	int currentbyte=currentbit/8;
	int currentbitinbyte=currentbit%8;
	int aux;
	int bitstogo=numberofbits;
	int bitindata;
	DWORD mask;

	bitindata=0;
	*data=0x0;
	while(bitstogo>0)
	{
		/* Bits left in current byte */
		aux=8-currentbitinbyte;
		/* Mask with size of bits left in byte or bits left in data */
		if(bitstogo<aux)
			aux=bitstogo;
		mask=(0x1<<aux)-1;
			
		*data = *data | (((buffer[currentbyte]>>currentbitinbyte) & mask)<<bitindata);
				
		/* Remaining data outside of current byte */
		if(currentbitinbyte+aux>=8)
		{
			currentbit+=aux;
			currentbyte++;
			currentbitinbyte=0;
			bitstogo-=aux;
			bitindata+=aux;
		}
		else /* No remaining data */
		{
			currentbitinbyte+=bitstogo;
			currentbit+=bitstogo;
			bitstogo=0;
		}
	}
	return currentbit;
}

