/**********************************************************************/
/***                                                                  */
/***        Author: Luis Eduardo Gamboa Guzmn                        */
/***        Last Updated: June 24th 2005                              */
/***                                                                  */
/***        Facultad de Ingeniera Elctrica                          */
/***        Universidad Michoacana de San Nicolas de Hidalgo          */
/***                                                                  */
/**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "wav.h"
#include "dsp.h"
#include "psychoacoustics.h"
#include "encode.h"
#include "types.h"
#include "huffman.h"
#include "arithmeticcoder.h"

/*#define COMPUTEGRAPHICS */


#define CLEANARRAY(x,N) {int macrok; for(macrok=0;macrok<N;macrok++) x[macrok]=0x0;}

#define BUFFERSIZE 32768
struct channel
{
	double *buffer;
	int currentpos;
	int extraindex;
};
typedef struct channel CHANNEL;

void encode(WAVFILE *wav,FILE *fout);
void initchannels(WAVFILE *wav,CHANNEL *left,CHANNEL *right);
void getblock(WAVFILE *wav,CHANNEL *left,CHANNEL *right,double *block,int channel,int blocksize,int overlap);
void printinfo(double *fftr,double *ffti,double *X,double *mask,int Nfreq,BYTE *quantizedr,UWORD length);

void decode(const char *strcuant,const char *strsalida);
void addblock(WAVFILE *wav,CHANNEL *left,CHANNEL *right,double *block,int channel,int blocksize,int overlap);
void flushbuffers(WAVFILE *wav,CHANNEL *left,CHANNEL *right);

int main(int numargs,char *args[])
{
	time_t start,end;
	//unsigned int cuenta;
	//double pruebad; DWORD prueba;
	WAVFILE wav;
	FILE *fout;

	if(numargs!=4)
	{
		printf("Syntax:\n %s input.wav output.legg output.wav\n",args[0]);
		return -1;
	}
	
	wav.fwav=fopen(args[1],"rb");
	time(&start);

	switch(readinfo(&wav))
	{
		case 0:
			printf("Sampling rate: %d\nChannels: %d\nBits per sample: %d\nSamples: %d, Length %fs\n\n",wav.samplerate,wav.channels,wav.bitspersample,wav.totalsamples,(float)(wav.totalsamples)/wav.samplerate);
			break;
		case -1:
			printf("%s is not a WAV file\n",args[1]);
			return -1;
		case -2:
			printf("%s is not a PCM file\n");
			return -1;
		case -3:
			printf("Sampling rate: %d\nChannels: %d\nBits per sample: %d\nSamples: %d\n",wav.samplerate,wav.channels,wav.bitspersample,wav.totalsamples);
			printf("Support limited to 44100Hz, 16 bits, Stereo\n");
			return -1;
		default:
			printf("Unknown error\n");
			return -1;
	}
	if(wav.totalsamples<BUFFERSIZE)
	{
		printf("File is too small, at least %d samples are required\n",BUFFERSIZE);
		return -1;
	}

	fout=fopen(args[2],"wb+");
	encode(&wav,fout);
	fclose(wav.fwav);
	fclose(fout);
	decode(args[2],args[3]);

	//dequantize(quantize(-40.77,0x3,5),0x3,5);

	/*pruebad=-0.9999;
	prueba=midtreadquantize(pruebad,3);
	printf("%e %x\n",pruebad,prueba);
	pruebad=midtreadvalue(prueba,3);
	printf("%e %x\n",pruebad,prueba);*/
	
	//huffmanencode("quantperc.legg",NULL,4);
	
	time(&end);
	printf("\n\nProcessing time: %d seconds\n",end-start);

	return 0;
}

void encode(WAVFILE *wav,FILE *fout)
{
	int i;
	int blocksize=512,Nfreq=blocksize/2;
	int maxfreq=wav->samplerate/2;
	int allocl,allocr;
	int numblocks;
	double *block=(double *)malloc(sizeof(double)*blocksize);
	double *X=(double *)malloc(sizeof(double)*Nfreq);
	double *fftr=(double *)malloc(sizeof(double)*Nfreq);
	double *ffti=(double *)malloc(sizeof(double)*Nfreq);
	double *mask=(double *)malloc(sizeof(double)*Nfreq);
	double *Xred=(double *)malloc(sizeof(double)*Nfreq);
	BYTE *quantizedl,*quantizedr;
	BYTE *codedl,*codedr;
	int lengthql,lengthqr,maxlength;
	double *window=(double *)malloc(sizeof(double)*blocksize);
	double *hanning=(double *)malloc(sizeof(double)*blocksize);
	struct hdrlegg hdr;
	struct hdrpckleggsize pckhdrsize;
	CHANNEL left,right;

	initchannels(wav,&left,&right);


	hdr.N=wav->totalsamples;
	hdr.mdctsize=Nfreq;
	hdr.sr=wav->samplerate;
	hdr.window=KBDW;
	hdr.channels=2;
	fwrite(&hdr,sizeof(struct hdrlegg),1,fout);
	
	hanningwindow(hanning,blocksize);
	switch(hdr.window)
	{
		case SINEW:
			sinewindow(window,blocksize);
			break;
		case KBDW:
			kbdwindow(window,blocksize,4.0);
			break;
		case VORBISW:
			vorbiswindow(window,blocksize);
			break;
		default:
			rectangularwindow(window,blocksize);
			break;
	}

	numblocks=(int)ceil(2.0*wav->totalsamples/blocksize)+1;

	/* Worst case scenario: each transform is quantized with a 32-bit quantizer */
	maxlength=sizeof(DWORD*)*Nfreq;
	quantizedl=(BYTE *)malloc(maxlength);
	quantizedr=(BYTE *)malloc(maxlength);
	codedr=(BYTE *)malloc(maxlength);
	codedl=(BYTE *)malloc(maxlength);

	initpsychoacoustic(Nfreq,maxfreq);
	for(i=0,allocl=0,allocr=0;i<numblocks;i++)
	{
		//printf("Block %d Left\n",i);
		getblock(wav,&left,&right,block,0,blocksize,blocksize/2);
		mdctblock(block,window,X,blocksize);
		fftblock(block,hanning,fftr,ffti,blocksize);

		computemask(fftr,ffti,Nfreq,maxfreq,mask);
		CLEANARRAY(quantizedl,maxlength);
		lengthql=quantizesignal(X,fftr,ffti,mask,Nfreq,quantizedl);/*+sizeof(struct hdrpckleggsize);*/

#ifdef COMPUTEGRAPHICS
		if(i==1)
			printinfo(fftr,ffti,X,mask,Nfreq,&quantizedl[sizeof(struct hdrpckleggsize)],(UWORD)lengthql);
#endif
		lengthql=arithmeticencode(&quantizedl[sizeof(struct hdrpckleggsize)],codedl,lengthql);
		allocl+=lengthql;
		
		getblock(wav,&left,&right,block,1,blocksize,blocksize/2);
		mdctblock(block,window,X,blocksize);
		fftblock(block,hanning,fftr,ffti,blocksize);
		computemask(fftr,ffti,Nfreq,maxfreq,mask);
		CLEANARRAY(quantizedr,maxlength);
		lengthqr=quantizesignal(X,fftr,ffti,mask,Nfreq,quantizedr);/*+sizeof(struct hdrpckleggsize);*/
		lengthqr=arithmeticencode(&quantizedr[sizeof(struct hdrpckleggsize)],codedr,lengthqr);
		allocr+=lengthqr;
		
		/* Writes encoded left channel */
		pckhdrsize.lengthofpacket=lengthql;
		fwrite(&pckhdrsize,sizeof(struct hdrpckleggsize),1,fout);
		fwrite(codedl,sizeof(BYTE),lengthql,fout);
		
		/* Writes encoded right channel */
		pckhdrsize.lengthofpacket=lengthqr;
		fwrite(&pckhdrsize,sizeof(struct hdrpckleggsize),1,fout);
		fwrite(codedr,sizeof(BYTE),lengthqr,fout);
		
		/*fwrite(quantizedl,sizeof(BYTE),lengthql,fout);
		fwrite(quantizedr,sizeof(BYTE),lengthqr,fout);*/
	}
	freepsychoacoustic();
	free(left.buffer);
	free(right.buffer);
	
	
	printf("Left channel, %d coefficients, %d bytes used\n",numblocks*blocksize/2,allocl);
	printf("Right channel, %d coefficients, %d bytes used\n",numblocks*blocksize/2,allocr);
	printf("File size: %d bytes %.4fkbps\n",allocl+allocr+sizeof(struct hdrlegg),(float)(allocl+allocr+sizeof(struct hdrlegg))*8.0/(1000.0*(float)wav->totalsamples/wav->samplerate));
	printf("End of Analysis, %d blocks\n\n",numblocks);

	free(quantizedl);
	free(quantizedr);
	free(X);
	free(fftr);
	free(ffti);
	free(mask);
	free(Xred);
	free(window);
	free(hanning);
}

void initchannels(WAVFILE *wav,CHANNEL *left,CHANNEL *right)
{
	int i;
	struct sample ssample;

	left->currentpos=0;
	right->currentpos=0;
	left->buffer=(double *)malloc(sizeof(double)*BUFFERSIZE);
	right->buffer=(double *)malloc(sizeof(double)*BUFFERSIZE);
	
	i=0;
	while(i<BUFFERSIZE && !readsample(wav,&ssample))
	{
		left->buffer[i]=ssample.wLeft/32768.0;
		right->buffer[i]=ssample.wRight/32768.0;
		i++;
	}
}

void getblock(WAVFILE *wav,CHANNEL *left,CHANNEL *right,double *block,int channel,int blocksize,int overlap)
{
	CHANNEL *chaux;
	struct sample ssample;
	int pos,i;

	switch(channel)
	{
		case 0:
			chaux=left;
			break;
		case 1:
			chaux=right;
			break;
	}
	/* Fill with 0's to the left */
	if(chaux->currentpos==0)
	{
		//printf("A la izq\n");
		for(i=0;i<overlap;i++)
			block[i]=0.0;
		memcpy(&block[i],chaux->buffer,sizeof(double)*(blocksize-overlap));
		chaux->currentpos+=blocksize-overlap;
	}
	else if(chaux->currentpos-overlap+blocksize<BUFFERSIZE)
	{ /* Data is in buffer */
		memcpy(block,&(chaux->buffer[chaux->currentpos-overlap]),sizeof(double)*blocksize);
		chaux->currentpos+=blocksize-overlap;
	}
	else
	{ /* Data is not in buffer, get from file */
		/* Save in buffer at least 1024 samples before currentpos*/
		/* currentpos of both channels should be the same */
		pos=left->currentpos;
		i=BUFFERSIZE-(pos-1024);
		memmove(left->buffer,&(left->buffer[pos-1024]),sizeof(double)*i);
		memmove(right->buffer,&(right->buffer[pos-1024]),sizeof(double)*i);
		left->currentpos=1024;
		right->currentpos=1024;

		while(i<BUFFERSIZE && !readsample(wav,&ssample))
		{
			left->buffer[i]=ssample.wLeft/32768.0;
			right->buffer[i]=ssample.wRight/32768.0;
			i++;
		}
		/* Fill with 0's to the right */
		if(i<BUFFERSIZE)
		{
			for(;i<BUFFERSIZE;i++)
			{
				left->buffer[i]=0.0;
				right->buffer[i]=0.0;
			}
		}
		memcpy(block,&(chaux->buffer[chaux->currentpos-overlap]),sizeof(double)*blocksize);
		chaux->currentpos+=blocksize-overlap;
	}
}


void printinfo(double *fftr,double *ffti,double *mdct,double *mask,int Nfreq,BYTE *quantized,UWORD length)
{
	struct hdrpckleggsize pckhdrsize;
	struct hdrpcklegg pckhdr;
	BYTE bitallocations[25];
	BYTE *buffer;
	double *daux,*daux1;
	int i;
	FILE *faux;


	daux=(double *)malloc(sizeof(double)*Nfreq);
	daux1=(double *)malloc(sizeof(double)*Nfreq);

	memcpy(&pckhdr,quantized,sizeof(struct hdrpcklegg));
	buffer=&quantized[sizeof(struct hdrpcklegg)];
	pckhdrsize.lengthofpacket=length;

	//memcpy(&pckhdrsize,quantized,sizeof(struct hdrpckleggsize));
	//memcpy(&pckhdr,&quantized[sizeof(struct hdrpcklegg)],sizeof(struct hdrpcklegg));
	/* Reads scale factors and mantisas for left channel */
	//buffer=&quantized[sizeof(struct hdrpckleggsize)+sizeof(struct hdrpcklegg)];
	bitallocations[0]=pckhdr.alloc0;
	bitallocations[1]=pckhdr.alloc1;
	bitallocations[2]=pckhdr.alloc2;
	bitallocations[3]=pckhdr.alloc3;
	bitallocations[4]=pckhdr.alloc4;
	bitallocations[5]=pckhdr.alloc5;
	bitallocations[6]=pckhdr.alloc6;
	bitallocations[7]=pckhdr.alloc7;
	bitallocations[8]=pckhdr.alloc8;
	bitallocations[9]=pckhdr.alloc9;
	bitallocations[10]=pckhdr.alloc10;
	bitallocations[11]=pckhdr.alloc11;
	bitallocations[12]=pckhdr.alloc12;
	bitallocations[13]=pckhdr.alloc13;
	bitallocations[14]=pckhdr.alloc14;
	bitallocations[15]=pckhdr.alloc15;
	bitallocations[16]=pckhdr.alloc16;
	bitallocations[17]=pckhdr.alloc17;
	bitallocations[18]=pckhdr.alloc18;
	bitallocations[19]=pckhdr.alloc19;
	bitallocations[20]=pckhdr.alloc20;
	bitallocations[21]=pckhdr.alloc21;
	bitallocations[22]=pckhdr.alloc22;
	bitallocations[23]=pckhdr.alloc23;
	bitallocations[24]=pckhdr.alloc24;
	dequantizesignal(bitallocations,buffer,pckhdrsize.lengthofpacket-sizeof(struct hdrpcklegg),daux);

	faux=fopen("graficas\\mdct","wb");
	for(i=0;i<Nfreq;i++)
		fprintf(faux,"%e %e\n",(double)(i*22050)/Nfreq,mdct[i]);
	fclose(faux);
	faux=fopen("graficas\\mdctspl","wb");
	daux1=mdctspl(mdct,Nfreq);
	for(i=0;i<Nfreq;i++)
		fprintf(faux,"%e %e\n",(double)(i*22050)/Nfreq,daux1[i]);
	fclose(faux);
	free(daux1);

	faux=fopen("graficas\\mdctquant","wb");
	for(i=0;i<Nfreq;i++)
		fprintf(faux,"%e %e\n",(double)(i*22050)/Nfreq,daux[i]);
	fclose(faux);
	faux=fopen("graficas\\mdctquantspl","wb");
	daux1=mdctspl(daux,Nfreq);
	for(i=0;i<Nfreq;i++)
		fprintf(faux,"%e %e\n",(double)(i*22050)/Nfreq,daux1[i]);
	fclose(faux);
	free(daux1);

	faux=fopen("graficas\\fftspl","wb");
	daux1=fftspl(fftr,ffti,Nfreq);
	for(i=0;i<Nfreq;i++)
		fprintf(faux,"%e %e\n",(double)(i*22050)/Nfreq,daux1[i]);
	fclose(faux);
	free(daux1);

	faux=fopen("graficas\\mask","wb");
	for(i=0;i<Nfreq;i++)
		fprintf(faux,"%e %e\n",(double)(i*22050)/Nfreq,mask[i]);
	fclose(faux);

	free(daux);
}

void decode(const char *strquant,const char *strout)
{
	FILE *fquant;
	struct hdrlegg hdr;
	struct hdrpcklegg pckhdr;
	struct hdrpckleggsize pckhdrsize;
	double *mdctleft,*mdctright;
	BYTE *buffer;
	BYTE bitallocations[25];
	int maxbufferlength;
	int mdctsize,N;
	double *bufferl,*bufferr;
	double *window;
	int i;
	WAVFILE wav;
	CHANNEL left,right;
	BYTE *codedr,*codedl;
	
	


	fquant=fopen(strquant,"rb");
	wav.fwav=fopen(strout,"wb+");

	fread(&hdr,sizeof(struct hdrlegg),1,fquant);

	N=hdr.N;
	mdctsize=hdr.mdctsize;

	printf("File info: %d %d %d %d ",N,mdctsize,hdr.sr,hdr.channels);
	wav.samplerate=hdr.sr;
	wav.bitspersample=16;
	wav.channels=hdr.channels;
	wav.totalsamples=hdr.N;
	wav.samplesread=0;

	left.currentpos=0;
	left.extraindex=0;
	left.buffer=(double *)malloc(sizeof(double)*BUFFERSIZE);
	right.currentpos=0;
	right.extraindex=0;
	right.buffer=(double *)malloc(sizeof(double)*BUFFERSIZE);

	bufferl=(double *)malloc(sizeof(double)*mdctsize*2);
	bufferr=(double *)malloc(sizeof(double)*mdctsize*2);
	mdctleft=(double *)malloc(sizeof(double)*mdctsize);
	mdctright=(double *)malloc(sizeof(double)*mdctsize);
	window=(double *)malloc(sizeof(double)*mdctsize*2);
	
	maxbufferlength=sizeof(DWORD)*mdctsize;
	buffer=(BYTE *)malloc(maxbufferlength);
	codedr=(BYTE *)malloc(maxbufferlength);
	codedl=(BYTE *)malloc(maxbufferlength);
	
	if(!bufferl || !bufferr || !mdctleft || !mdctright || !window)
	{
		printf("Error, out of memory\n");
		exit(-1);
	}
	
	switch(hdr.window)
	{
		case SINEW:
			sinewindow(window,mdctsize*2);
			printf("sine ");
			break;
		case KBDW:
			kbdwindow(window,mdctsize*2,4.0);
			printf("kbd ");
			break;
		case VORBISW:
			vorbiswindow(window,mdctsize*2);
			printf("vorbis ");
			break;
		default:
			rectangularwindow(window,mdctsize*2);
			printf("rectangular ");
			break;
	}

	printf("\n");
	writeinfo(&wav);
	
	initpsychoacoustic(mdctsize,hdr.sr/2);
	i=0;
	while(fread(&pckhdrsize,sizeof(struct hdrpckleggsize),1,fquant))
	{
		fread(codedl,sizeof(BYTE),pckhdrsize.lengthofpacket,fquant);
		/* Decodes packet */
		pckhdrsize.lengthofpacket=arithmeticdecode(codedl,buffer,pckhdrsize.lengthofpacket);
		/* Reads scale factors and mantisas for left channel */
		memcpy(&pckhdr,buffer,sizeof(struct hdrpcklegg));
		memmove(buffer,&buffer[sizeof(struct hdrpcklegg)],pckhdrsize.lengthofpacket-sizeof(struct hdrpcklegg));
		/*fread(&pckhdr,sizeof(struct hdrpcklegg),1,fquant);
		fread(buffer,sizeof(BYTE),pckhdrsize.lengthofpacket-sizeof(struct hdrpcklegg),fquant);*/
		bitallocations[0]=pckhdr.alloc0;
		bitallocations[1]=pckhdr.alloc1;
		bitallocations[2]=pckhdr.alloc2;
		bitallocations[3]=pckhdr.alloc3;
		bitallocations[4]=pckhdr.alloc4;
		bitallocations[5]=pckhdr.alloc5;
		bitallocations[6]=pckhdr.alloc6;
		bitallocations[7]=pckhdr.alloc7;
		bitallocations[8]=pckhdr.alloc8;
		bitallocations[9]=pckhdr.alloc9;
		bitallocations[10]=pckhdr.alloc10;
		bitallocations[11]=pckhdr.alloc11;
		bitallocations[12]=pckhdr.alloc12;
		bitallocations[13]=pckhdr.alloc13;
		bitallocations[14]=pckhdr.alloc14;
		bitallocations[15]=pckhdr.alloc15;
		bitallocations[16]=pckhdr.alloc16;
		bitallocations[17]=pckhdr.alloc17;
		bitallocations[18]=pckhdr.alloc18;
		bitallocations[19]=pckhdr.alloc19;
		bitallocations[20]=pckhdr.alloc20;
		bitallocations[21]=pckhdr.alloc21;
		bitallocations[22]=pckhdr.alloc22;
		bitallocations[23]=pckhdr.alloc23;
		bitallocations[24]=pckhdr.alloc24;
		dequantizesignal(bitallocations,buffer,pckhdrsize.lengthofpacket-sizeof(struct hdrpcklegg),mdctleft);
		
		/* Reads packet header for right channel */
		fread(&pckhdrsize,sizeof(struct hdrpckleggsize),1,fquant);
		fread(codedr,sizeof(BYTE),pckhdrsize.lengthofpacket,fquant);
		/* Decodes packet */
		pckhdrsize.lengthofpacket=arithmeticdecode(codedr,buffer,pckhdrsize.lengthofpacket);
		/* Reads scale factors and mantisas for left channel */
		memcpy(&pckhdr,buffer,sizeof(struct hdrpcklegg));
		memmove(buffer,&buffer[sizeof(struct hdrpcklegg)],pckhdrsize.lengthofpacket-sizeof(struct hdrpcklegg));
		/*fread(&pckhdr,sizeof(struct hdrpcklegg),1,fquant);
		fread(buffer,sizeof(BYTE),pckhdrsize.lengthofpacket-sizeof(struct hdrpcklegg),fquant);*/
		/* Scale factors and mantisas for right channel */
		bitallocations[0]=pckhdr.alloc0;
		bitallocations[1]=pckhdr.alloc1;
		bitallocations[2]=pckhdr.alloc2;
		bitallocations[3]=pckhdr.alloc3;
		bitallocations[4]=pckhdr.alloc4;
		bitallocations[5]=pckhdr.alloc5;
		bitallocations[6]=pckhdr.alloc6;
		bitallocations[7]=pckhdr.alloc7;
		bitallocations[8]=pckhdr.alloc8;
		bitallocations[9]=pckhdr.alloc9;
		bitallocations[10]=pckhdr.alloc10;
		bitallocations[11]=pckhdr.alloc11;
		bitallocations[12]=pckhdr.alloc12;
		bitallocations[13]=pckhdr.alloc13;
		bitallocations[14]=pckhdr.alloc14;
		bitallocations[15]=pckhdr.alloc15;
		bitallocations[16]=pckhdr.alloc16;
		bitallocations[17]=pckhdr.alloc17;
		bitallocations[18]=pckhdr.alloc18;
		bitallocations[19]=pckhdr.alloc19;
		bitallocations[20]=pckhdr.alloc20;
		bitallocations[21]=pckhdr.alloc21;
		bitallocations[22]=pckhdr.alloc22;
		bitallocations[23]=pckhdr.alloc23;
		bitallocations[24]=pckhdr.alloc24;
		dequantizesignal(bitallocations,buffer,pckhdrsize.lengthofpacket-sizeof(struct hdrpcklegg),mdctright);

		imdctblock(bufferl,window,mdctleft,mdctsize*2);
		imdctblock(bufferr,window,mdctright,mdctsize*2);
		
		addblock(&wav,&left,&right,bufferl,0,mdctsize*2,mdctsize);
		addblock(&wav,&left,&right,bufferr,1,mdctsize*2,mdctsize);
		i++;
	}
	flushbuffers(&wav,&left,&right);
	freepsychoacoustic();

	printf("End of Synthesis, %d samples\n\n\n",i);
	fclose(wav.fwav);

	free(window);
	free(bufferl);
	free(bufferr);
	free(mdctleft);
	free(mdctright);

	free(left.buffer);
	free(right.buffer);

	fclose(fquant);
}

/* 
	addblock: overlaps and adds imdct blocks of the left and right channel,
	when the buffer is full it flushes the buffer contents to wav. Can only
	manage 2 channels.

  parameters:
	channel: 0 is left, 1 is right
	blocksize: number of samples in the imdct block
	overlap: number of samples that overlap with the previous block
*/
void addblock(WAVFILE *wav,CHANNEL *left,CHANNEL *right,double *block,int channel,int blocksize,int overlap)
{
	CHANNEL *chaux;
	int pos,i,j;

	switch(channel)
	{
		case 0:
			chaux=left;
			break;
		case 1:
			chaux=right;
			break;
	}
	/* Assume 0's to the left */
	if(chaux->currentpos==0)
	{
		//printf("A la izq\n");
		for(i=0,j=overlap;j<blocksize;i++,j++)
			chaux->buffer[i]=block[j];
		for(;i<BUFFERSIZE;i++)
			chaux->buffer[i]=0.0;
		chaux->currentpos+=blocksize-overlap;
		chaux->extraindex=0;
	}
	else if(chaux->currentpos-overlap+blocksize<BUFFERSIZE)
	{ /* Buffer is not full */
		for(i=chaux->currentpos-overlap,j=0;j<blocksize;i++,j++)
			chaux->buffer[i]+=block[j];
		chaux->extraindex+=overlap;
		chaux->currentpos+=blocksize-overlap;
	}
	else
	{ /* Buffer is full, write to file */
		i=0;
		while(i<left->extraindex && i<right->extraindex)
		{
			writesample(wav,left->buffer[i],right->buffer[i]);
			i++;
		}

		pos=i;
		i=left->currentpos-pos;
		left->extraindex-=pos;
		memmove(left->buffer,&(left->buffer[pos]),sizeof(double)*i);
		left->currentpos-=pos;

		i=right->currentpos-pos;
		right->extraindex-=pos;
		memmove(right->buffer,&(right->buffer[pos]),sizeof(double)*i);
		right->currentpos-=pos;

		/* Fill with 0's to the right */
		for(i=left->currentpos;i<BUFFERSIZE;i++)
			left->buffer[i]=0.0;
		for(i=right->currentpos;i<BUFFERSIZE;i++)
			right->buffer[i]=0.0;

		for(i=chaux->currentpos-overlap,j=0;j<blocksize;i++,j++)
			chaux->buffer[i]+=block[j];
		chaux->extraindex+=overlap;
		chaux->currentpos+=blocksize-overlap;
	}
}

void flushbuffers(WAVFILE *wav,CHANNEL *left,CHANNEL *right)
{
	int i;
	if(left->extraindex!=right->extraindex)
		printf("Aguas!!! %d %d \n",left->extraindex,right->extraindex);
	for(i=0;i<left->extraindex;i++)
		writesample(wav,left->buffer[i],right->buffer[i]);
}

